<?php

/**
 * Allows you to easily package RPG Maker MV plugins.
 */
class Rmmvpackager {
    private $_packageFolder;
    private $_defaultNamespace;
    
    /**
     * Constructor.
     * @param array|string $packageFolder The folder where the packager files reside.
     * @param string $defaultNamespace The default namespace.
     */
    public function __construct($packagerFolder = '', $defaultNamespace = 'Rmmv.Package') {
        if (is_array($packagerFolder)) {
            $this->_packagerFolder = rtrim($packagerFolder['packageFolder'], '/') . '/';
            $this->_defaultNamespace = $packagerFolder['defaultNamespace'];
        } else {
            $this->_packagerFolder = rtrim($packagerFolder, '/') . '/';
            $this->_defaultNamespace = $defaultNamespace;
        }
    }
    
    /**
     * Packages the plugins.
     * @param array $plugins An array of plugin filenames.
     * @param array $options Additional options.
     * @returns An array containing a script and a filename.
     */
    public function package($plugins, $options = []) {
        $options['plugins'] = $plugins;
        $this->set_defaults($options);
        
        $script = preg_replace('/(\r\n|\r)/', "\n", file_get_contents($this->_packagerFolder . 'main.js'));
        
        $script = $this->process_script($script, $options);
        
        $filename = !empty($options['filename']) ? $options['filename'] : (str_replace('.', '_', $options['namespace']) . '_<HASH>');
        $filename = str_replace('<HASH>', md5($script), $filename);
        
        preg_match('/\* (\<FILENAME\>\.js[ ]+) \*/', $script, $matches);
        
        $script = str_replace($matches[0], '* ' . str_pad($filename . '.js', strlen($matches[1]), ' ') . ' *', $script);
        $script = str_replace('/* FILENAME */', $filename, $script);
        
        return [
            'script' => $script,
            'filename'  => $filename,
        ];
    }
    
    private function set_defaults(&$post) {
        $post['namespace'] = empty($post['namespace']) ? $this->_defaultNamespace : $post['namespace'];
    }
    
    private function process_script($script, $post) {
        $script = $this->set_namespace_script($script, $post);
        $script = $this->set_scripts($script, $post);
        
        return $script;
    }
    
    private function set_namespace_script($script, $post) {
        $namespace_segments = explode('.', $post['namespace']);
        
        preg_match('/([ ]+)\/\* == NAMESPACE == \*\//', $script, $matches);
        
        $ns = 'window';
        $text = '';
        while (!empty($namespace_segments)) {
            $ns.= '.' . array_shift($namespace_segments);
            $text.= "$ns = $ns || {}\n";
        }
        $script = str_replace($matches[0], $this->indent(trim($text, "\n"), $matches[1]), $script);
        $script = str_replace('<NAMESPACE>', $post['namespace'], $script);
        $script = preg_replace('/\/\* NAMESPACE \*\//', $post['namespace'] . '.', $script);
        
        return $script;
    }
    
    private function indent($script, $indent = '') {
        $lines = explode("\n", $script);
        if (!is_string($indent)) {
            $indent = str_pad('', $indent, ' ');
        }

        for ($idx = 0; $idx < count($lines); $idx++) {
            $lines[$idx] = $indent . $lines[$idx];
        }
        return implode("\n", $lines);
    }
    
    private function set_scripts($script, $post) {
        $plugins = [];
        
        if (!empty($post['plugin'])) {
            foreach (array_keys($post['plugin']) as $plugin) {
                $file_base = $this->_packagerFolder . 'plugins/' . $plugin;
                $file_script = $file_base . '.js';
                $file_data = $file_base . '.json';
                $file_help = $file_base . '.txt';
                
                if (file_exists($file_script)) {
                    $plugin_data = [];
                    if (file_exists($file_data)) {
                        $plugin_data = json_decode(file_get_contents($file_data), true);
                        if (empty($plugin_data)) {
                            $plugin_data = [];
                        }
                    }
                    $plugin_script = preg_replace('/(\r\n|\r)/', "\n", file_get_contents($file_script));
                    $plugin_data['script'] = $this->parse_script($plugin_script, $post);
                    
                    if (file_exists($file_help)) {
                        $text = preg_replace('/(\r\n|\r)/', "\n", file_get_contents($file_help));
                        $text = str_replace('<NAMESPACE>', $post['namespace'], $text);
                        $plugin_data['help'] = $text;
                    }
                    
                    $plugins[] = $plugin_data;
                }
            }
        }
        
        $script = $this->set_game_interpreter_overrides($script, $post, $plugins);
        
        preg_match_all('/([ ]+)\/\* == (.+?) == \*\//', $script, $matches, PREG_SET_ORDER);

        foreach ($matches as $match) {
            $indents = $match[1];
            $section_name = $match[2];
            $section_scripts = [];
            $section = '';
            foreach ($plugins as $plugin) {
                if (!empty($plugin['script'][$section_name])) {
                    $section_scripts[] = $plugin['script'][$section_name];
                }
            }
            if (!empty($section_scripts)) {
                $section = implode("\n\n", $section_scripts);
                $section = $this->indent($section, $indents);
            }
            $script = str_replace($match[0], $section, $script);
        }
        
        $script = $this->set_parameters($script, $plugins);
        $script = $this->set_package_list($script, $plugins);
        $script = $this->set_help_text($script, $plugins);
        $script = $this->set_plugin_overrides($script, $plugins);
        
        $script = preg_replace('/[ ]+\n/', "\n", $script);
        $script = preg_replace('/\n{3,}/', "\n\n", $script);
        $script = preg_replace('/( \*\n){2,}/', " *\n", $script);
        
        return $script;
    }
    
    private function set_game_interpreter_overrides($script, $post, $plugins) {
        preg_match('/([ ]+\* )- (!== GAME_INTERPRETER ==[ ]+) -/', $script, $comment_match);
        $comment_indent = $comment_match[1];
        $comment_width = strlen($comment_match[2]);
        $methods = [];
        $methods_comment = [];
        $comment = '';
        $text = '';
        foreach ($plugins as $plugin) {
            if (!empty($plugin['methods'])) {
                foreach ($plugin['methods'] as $method) {
                    $methods_comment[] = '- ' . str_pad('Game_Interpreter.prototype.' . $method . ' (New)', $comment_width, ' ') . ' -';
                    $methods[] = 'Game_Interpreter.prototype.' . $method . ' = ' . $post['namespace'] . '.' . $method . ';';
                }
            }
        }
        if (!empty($methods_comment)) {
            $comment = implode("\n", $methods_comment);
        }
        if (!empty($methods)) {
            $text = implode("\n", $methods);
        }
        
        preg_match('/([ ]+)\/\* == GAME_INTERPRETER == \*\//', $script, $matches);
        $script = str_replace($matches[0], $this->indent(trim($text, "\n"), $matches[1]), $script);
        $script = str_replace($comment_match[0], $this->indent(trim($comment, "\n"), $comment_indent), $script);
        
        return $script;
    }
    
    private function parse_script($script, $post) {
        $script = preg_replace('/\/\* NAMESPACE \*\//', $post['namespace'] . '.', $script);
        preg_match_all('/\/\* == (.+?) == \*\//', $script, $matches, PREG_SET_ORDER | PREG_OFFSET_CAPTURE);
        
        $sections = [];
        
        for ($idx = 0; $idx < count($matches); $idx++) {
            $offset = $matches[$idx][0][1] + strlen($matches[$idx][0][0]);
            $nextOffset = strlen($script);
            if ($idx < count($matches) - 1) {
                $nextOffset = $matches[$idx + 1][0][1];
            }
            $section = trim(substr($script, $offset, $nextOffset - $offset), "\n");
            $sections[$matches[$idx][1][0]] = $section;
        }
        
        return $sections;
    }
    
    private function set_parameters($script, $plugins) {
        $params = [];
        foreach ($plugins as $plugin) {
            if (!empty($plugin['params'])) {
                foreach ($plugin['params'] as $param) {
                    $vars = [];
                    foreach ($param as $key => $val) {
                        $vars[] = '@' . $key . ' ' . $val;
                    }
                    if (!empty($vars)) {
                        $params[] = implode("\n", $vars);
                    }
                }
            }
        }
        $text = '';
        
        if (!empty($params)) {
            $text = $this->indent(implode("\n\n", $params), ' * ');
            $text = preg_replace('/ \* \n/', " *\n", $text) . "\n";
        }
        return str_replace(" * !== PLUGIN PARAMETERS ==\n", $text, $script);
    }
    
    private function set_package_list($script, $plugins) {
        $plugin_str = '';
        $plugins_text = [];
        foreach ($plugins as $plugin) {
            if (!empty($plugin['name'])) {
                $plugins_text[] = '* ' . $plugin['name'];
            }
        }
        if (!empty($plugins_text)) {
            $plugin_str = $this->indent(implode("\n", $plugins_text), ' * ') . "\n";
        }
        
        $script = str_replace(" * !== PLUGIN LIST ==\n", $plugin_str, $script);
        
        return $script;
    }
    
    private function set_help_text($script, $plugins) {
        $help = '';
        $help_text = [];
        foreach ($plugins as $plugin) {
            if (!empty($plugin['help'])) {
                $text = str_pad('', 76, '-') . "\n";
                $text.= '- ' . str_pad($plugin['name'], 72, ' ') . " -\n";
                $text.= str_pad('', 76, '-') . "\n\n";
                $text.= $plugin['help'];
                $help_text[] = $text;
            }
        }
        if (!empty($help_text)) {
            $help = $this->indent(implode("\n\n", $help_text), ' * ');
            $help = preg_replace('/ \* \n/', " *\n", $help) . "\n";
        }
        
        $script = str_replace(" * !== PLUGIN USAGE ==\n", $help, $script);
        
        return $script;
    }
    
    private function set_plugin_overrides($script, $plugins) {
        $overrides = [];
        foreach ($plugins as $plugin) {
            if (!empty($plugin['overrides'])) {
                foreach ($plugin['overrides'] as $override) {
                    if (!in_array($override, $overrides)) {
                        $overrides[] = $override;
                    }
                }
            }
        }
        $text = 'This plugin does not overwrite default functionality.';
        if (!empty($overrides)) {
            $text = "This plugin overwrites default functionality. Make sure you check whether or\n";
            $text.= "not the plugin is compatible with other plugins by checking which functions\n";
            $text.= "they overwrite. Below is the list of methods it overwrites:\n\n";
            sort($overrides);
            foreach ($overrides as $override) {
                $text.= "* $override\n";
            }
            
            $text = $this->indent($text, ' * ');
            $text = preg_replace('/ \* \n/', " *\n", $text) . "\n";
        }
        
        $script = str_replace(" * !== PLUGIN OVERRIDES ==\n", $text, $script);
        return $script;
    }
}
