/******************************************************************************
 * <FILENAME>.js                                                              *
 ******************************************************************************
 * By G.A.M. Kertopermono, a.k.a. GaryCXJk                                    *
 ******************************************************************************
 * License: ISC                                                               *
 ******************************************************************************
 * Copyright (c) 2017, G.A.M. Kertopermono                                    *
 *                                                                            *
 * Permission to use, copy, modify, and/or distribute this software for any   *
 * purpose with or without fee is hereby granted, provided that the above     *
 * copyright notice and this permission notice appear in all copies.          *
 *                                                                            *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES   *
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF           *
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR    *
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES     *
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN      *
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR *
 * IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.                *
 ******************************************************************************/

/*:
 * @plugindesc Contains a set of plugins selected by the user.
 * @author G.A.M. Kertopermono
 *
 * @param general
 * @text General settings
 *
 * @param general.override_pluginmanager
 * @text Override PluginManager.parameters
 * @desc Override the original parameters function?
 * @parent general
 * @type boolean
 * @on Override
 * @off Don't override
 * @default false
 *
 * @param general.append_to_interpreter
 * @text Modify Game_Interpreter
 * @desc Add functions to Game_Interpreter that manipulate it?
 * @parent general
 * @type boolean
 * @on Yes
 * @off No
 * @default false
 *
 * !== PLUGIN PARAMETERS ==
 *
 * @help
 * ============================================================================
 * = About                                                                    =
 * ============================================================================
 *
 * This contains the following selected packages:
 *
 * !== PLUGIN LIST ==
 *
 * ============================================================================
 * = Usage                                                                    =
 * ============================================================================
 *
 * Below the usage of each plugin is explained.
 *
 * ----------------------------------------------------------------------------
 * - Default package functionality                                            -
 * ----------------------------------------------------------------------------
 *
 * The following functions are added by default, since the other packages may
 * need this functionality.
 *
 * ---
 *
 * <NAMESPACE>.getParameters(pluginName, defaultParameters)
 *
 * This retrieves the parameters for the current plugin. It essentially is an
 * extension to the existing PluginManager function. First, it calls
 * PluginManager.parameters to retrieve the parameters in the old-fashioned
 * way. If it can't find the parameters, it falls back to
 * document.currentScript, which essentially gets the current script file
 * name. This does not work on Internet Explorer, so as a final resort it
 * checks all settings for each plugin and compares it to the object given in
 * defaultParameters. If it finds a complete match, it'll return these
 * parameters. Finally, if it still fails to find the plugin settings, it will
 * use the defaultParameters object as the parameters.
 *
 * Arguments:
 *
 * pluginName        - The name of the plugin
 * defaultParameters - The default parameters
 *
 * Returns:
 *
 * An object containing every parameter, or defaultParameters if it can't find
 * it.
 *
 * = Changelog =
 *
 * 1.1
 * ---
 * * Fixed an issue with plugin parameters not inheriting missing defaults
 *
 * 1.0
 * ---
 * * Initial release
 *
 * !== PLUGIN USAGE ==
 *
 * ============================================================================
 * = Compatibility                                                            =
 * ============================================================================
 *
 * !== PLUGIN OVERRIDES ==
 *
 * ============================================================================
 * = License                                                                  =
 * ============================================================================
 *
 * Copyright (c) 2017, G.A.M. Kertopermono
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR
 * IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * ============================================================================
 */

// Sets variables for the dependency check
var Imported = Imported || {};

(function() {
    /* == NAMESPACE == */

    /* ------------------------------------------------------------------------
     * - PRIVATE VARIABLES                                                    -
     * ------------------------------------------------------------------------
     */

    var _coreExists = {};

    /* == PRIVATE VARIABLES == */

    /* ------------------------------------------------------------------------
     * - PRIVATE FUNCTIONS                                                    -
     * ------------------------------------------------------------------------ 
     */

    /* Checks plugin content */
    var _checkPluginContent = function(parameters) {
        for(var prop in parameters) {
            if(parameters.hasOwnProperty(prop)) {
                return true;
            }
        }
        return false;
    }

    var _deepMergeObject = function(object, otherObject) {
        for(var prop in otherObject) {
          if (otherObject.hasOwnProperty(prop)) {
            if (object.hasOwnProperty(prop)) {
              if (typeof object[prop] === 'object' && typeof otherObject[prop] === 'object') {
                return _deepMergeObject(object[prop], otherObject[prop]);
                continue;
              }
            }
          }
          object[prop] = otherObject[prop];
        }
        return object;
    }

    /* == PRIVATE FUNCTIONS == */

    /* ------------------------------------------------------------------------
     * - ALIASES                                                              -
     * ------------------------------------------------------------------------
     */
    
    var _old = {};
    
    _old.pluginManagerParameters = PluginManager.parameters;

    /* == ALIASES == */
    
    /* NAMESPACE */old = _old;

    /* ------------------------------------------------------------------------
     * - Plugin methods                                                       -
     * ------------------------------------------------------------------------
     */

    /**
     * Gets the current script's parameters.
     * When the script's name is renamed, it tries various fallbacks before
     * defaulting to the default parameters given.
     */
    /* NAMESPACE */getParameters = function(pluginName, defaultParams) {
        // Use the default RPG Maker MV plugin manager
        var parameters = /* NAMESPACE */old.pluginManagerParameters.call(PluginManager, pluginName);
        if(_checkPluginContent(parameters)) {
            return _deepMergeObject(parameters, defaultParams);
        }
        
        // Use currentScript (doesn't work on Internet Explorer)
        if(document.currentScript) {
            var scriptName = document.currentScript.src;
            scriptName = scriptName.substr(scriptName.indexOf('/plugins/') + 9);
            scriptName = scriptName.substr(0, scriptName.lastIndexOf('.js'));
            parameters = /* NAMESPACE */old.pluginManagerParameters.call(PluginManager, scriptName);
            return _deepMergeObject(parameters, defaultParams);
        }
        
        // Iterate through each defined script
        for(var idx = 0; idx < $plugins.length; idx++) {
            var plugin = $plugins[idx];
            var params = plugin.parameters;
            
            // If the description contains the plugin name, this is the guy
            if(plugin.description.indexOf('<' + pluginName + '>') > -1) {
              return _deepMergeObject(params, defaultParams);
            }
            
            /* Check each parameter of the plugin, if all are present in the
               defaultParams parameter, this is the script */
            var hasFound = true;
            if(defaultParams) {
                var propNames = Object.keys(defaultParams);
                for(var idx = 0; idx < propNames.length; idx++) {
                    if(!params.hasOwnProperty(propNames[idx])) {
                        hasFound = false;
                        break;
                    }
                }
            }
            if(hasFound) {
              return _deepMergeObject(params, defaultParams);
            }
        }
        
        // Return the default parameters
        return defaultParams;
    }
    
    /* NAMESPACE */getParameters.version = '1.1';
    
    /* == PLUGIN METHODS == */
    
    (function() {
        var defaultParameters = {};
        
        defaultParameters['general.override_pluginmanager'] = 'false';
        defaultParameters['general.append_to_interpreter'] = 'false';

        /* == PLUGIN DEFAULT PARAMETERS == */

        var parameters = /* NAMESPACE */getParameters('/* FILENAME */', defaultParameters);
        
        /* Override PluginManager.parameters
         * Overrides the original PluginManager.parameters function with the
         * packaged one.
         */
        if(parameters['general.override_pluginmanager'] !== 'false') {
            PluginManager.parameters = /* NAMESPACE */getParameters;
        }
        
        /* == PLUGIN OVERRIDES == */

        /* --------------------------------------------------------------------
         * - !== GAME_INTERPRETER ==                                          -
         * --------------------------------------------------------------------
         */
        if(parameters['general.append_to_interpreter'] !== 'false') {
            /* == GAME_INTERPRETER == */
        }
    })();
})();
